import copy
import random
import sys
import time

from tala.testing.interactions.base_test import BaseInteractionTest


class MissingTurnsException(Exception):
    pass


class EnduranceInteractionTest(BaseInteractionTest):
    def __init__(self, all_tests, duration):
        self._all_tests = list(all_tests)
        self._start_time = time.time()
        self._duration = duration
        self._is_requested_to_stop = False

    @property
    def _end_time(self):
        if self._duration <= 0:
            return sys.maxsize
        else:
            return self._start_time + self._duration

    @property
    def filename(self):
        return self.__class__.__name__

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def turns(self):
        while not self._is_requested_to_stop and time.time() < self._end_time:
            next_test = self._prepare_next_test()
            for turn in next_test.turns:
                yield turn

    def _prepare_next_test(self):
        next_test = self._randomly_sampled_test()
        self._remove_first_turn_if_type_is_system(next_test)
        if not next_test.turns:
            raise MissingTurnsException("Expected turns but found none in test '%s' of '%s'" %
                                        (next_test.name, next_test.filename))
        return next_test

    @staticmethod
    def _remove_first_turn_if_type_is_system(next_test):
        first_turn = next_test.turns[0]
        if first_turn.is_system_output_turn:
            next_test.turns.pop(0)

    def _randomly_sampled_test(self):
        random_test = random.choice(self._all_tests)
        if not random_test.turns:
            raise MissingTurnsException("Expected turns but found none in test '%s' of '%s'" %
                                        (random_test.name, random_test.filename))
        return copy.copy(random_test)

    def stop(self):
        self._is_requested_to_stop = True



from tala.model.stack import Stack
from tala.model import plan_item
from tala.model.semantic_object import SemanticObject


class UnableToDetermineOntologyException(Exception):
    pass


class Plan(Stack, SemanticObject):
    def __init__(self, content=None):
        if content is None:
            content = set()
        Stack.__init__(self, content=content)
        SemanticObject.__init__(self)

    def is_ontology_specific(self):
        if not all([proposition.is_ontology_specific() for proposition in self.content]):
            return False
        if not self._ontology_names:
            return False
        return self._are_all_propositions_from_same_ontology()

    def _are_all_propositions_from_same_ontology(self):
        return len(self._ontology_names) < 2

    @property
    def ontology_name(self):
        if not self._ontology_names:
            raise UnableToDetermineOntologyException("Expected semantic content but found none")

        if self.is_ontology_specific():
            return self._ontology_names.pop()

        message = "Expected all propositions %s\n\nin ontology %s\n\nbut they're from %s" % (
            self.content, self.content[0].ontology_name, self._ontology_names
        )
        raise UnableToDetermineOntologyException(message)

    @property
    def _ontology_names(self):
        ontology_names = set(proposition.ontology_name for proposition in self.content)
        return ontology_names

    def has_semantic_content(self):
        return True

    def __iter__(self):
        flattened_items = self._flatten(self.content)
        return flattened_items.__iter__()

    def _flatten(self, items):
        result = []
        for item in items:
            if item.type_ == plan_item.TYPE_IF_THEN_ELSE:
                if item.consequent:
                    result.extend(self._flatten(item.get_consequent()))
                if item.alternative:
                    result.extend(self._flatten(item.get_alternative()))
            else:
                result.append(item)
        return result

    def remove(self, item_to_remove):
        items_to_remove = []
        for item in self.content:
            if item == item_to_remove:
                items_to_remove.append(item)
            elif item.type_ == plan_item.TYPE_IF_THEN_ELSE:
                self.remove_nested(item_to_remove, item)
        for item in items_to_remove:
            self.content.remove(item)

    def remove_nested(self, to_remove, item):
        if to_remove in item.consequent:
            item.consequent.remove(to_remove)
        if to_remove in item.alternative:
            item.alternative.remove(to_remove)
        for sub_item in item.alternative + item.consequent:
            if sub_item.type_ == plan_item.TYPE_IF_THEN_ELSE:
                self.remove_nested(to_remove, sub_item)

    def get_questions_in_plan_without_feature_question(self):
        for item in self:
            if item.type_ in plan_item.QUESTION_TYPES:
                question = item.content
                yield question

    def __str__(self):
        return "Plan(%s)" % self.content

    def __repr__(self):
        return "%s%s" % (self.__class__.__name__, (self.content))


class InvalidPlansException(Exception):
    pass



import sys
import unittest

from tala.testing.interactions.file import InteractionTestingFile
from tala.testing.interactions.testcase import InteractionTestingTestCase
from tala.testing.endurance.named_test import EnduranceInteractionTest


class EnduranceTestRunner(object):
    FAILURE_EXIT_CODE = 5

    def __init__(self, test_files, duration, environment_or_url):
        self._tests = self._all_tests_from_files(test_files)
        self._duration = duration
        self._environment_or_url = environment_or_url
        self._endurance_interaction_test = EnduranceInteractionTest(self._tests, self._duration)

    @staticmethod
    def _all_tests_from_files(test_files):
        test_files = [InteractionTestingFile.from_path(test_file) for test_file in test_files]
        for test_file in test_files:
            for test in test_file.tests:
                yield test

    def run(self):
        test_case = InteractionTestingTestCase(self._endurance_interaction_test,
                                               self._environment_or_url)
        suite = unittest.TestSuite()
        suite.addTest(test_case)
        successful = self._run(suite)
        if not successful:
            sys.exit(self.FAILURE_EXIT_CODE)

    @staticmethod
    def _run(suite):
        runner = unittest.TextTestRunner()
        result = runner.run(suite)
        successful = result.wasSuccessful()
        return successful

    def stop(self):
        self._endurance_interaction_test.stop()



from tala.model.semantic_object import SemanticObject

HAS_VALUE = "HasValue"
IS_TRUE = "IsTrue"
HAS_SHARED_VALUE = "HasSharedValue"
HAS_PRIVATE_VALUE = "HasPrivateValue"
HAS_SHARED_OR_PRIVATE_VALUE = "HasSharedOrPrivateValue"
IS_SHARED_COMMITMENT = "IsSharedCommitment"
IS_PRIVATE_BELIEF = "IsPrivateBelief"
IS_PRIVATE_BELIEF_OR_SHARED_COMMITMENT = "IsPrivateBeliefOrSharedCommitment"
QUERY_HAS_MORE_ITEMS = "QueryHasMoreItems"

SEMANTIC_OBJECT_TYPE = "condition"


def create_condition(type_, parameter):
    if type_ == HAS_SHARED_VALUE:
        return HasSharedValue(parameter)
    if type_ == HAS_PRIVATE_VALUE:
        return HasPrivateValue(parameter)
    if type_ == HAS_SHARED_OR_PRIVATE_VALUE:
        return HasSharedOrPrivateValue(parameter)
    if type_ == IS_SHARED_COMMITMENT:
        return IsSharedCommitment(parameter)
    if type_ == IS_PRIVATE_BELIEF:
        return IsPrivateBelief(parameter)
    if type_ == IS_PRIVATE_BELIEF_OR_SHARED_COMMITMENT:
        return IsPrivateBeliefOrSharedCommitment(parameter)
    if type_ == QUERY_HAS_MORE_ITEMS:
        return QueryHasMoreItems(parameter)
    if type_ == HAS_VALUE:
        return HasValue(parameter)
    if type_ == IS_TRUE:
        return IsTrue(parameter)


class Condition(SemanticObject):
    def __init__(self, type_):
        self._type = type_

    @property
    def condition_type(self):
        return self._type

    def is_has_value_condition(self):
        return self.condition_type in [HAS_VALUE, HAS_SHARED_VALUE, HAS_PRIVATE_VALUE, HAS_SHARED_OR_PRIVATE_VALUE]

    def is_is_true_condition(self):
        return self.condition_type in [
            IS_TRUE, IS_SHARED_COMMITMENT, IS_PRIVATE_BELIEF, IS_PRIVATE_BELIEF_OR_SHARED_COMMITMENT
        ]

    def is_true_given_proposition_set(self, _):
        return False

    def __eq__(self, other):
        try:
            return self.condition_type == other.condition_type
        except AttributeError:
            return False

    def __repr__(self):
        return "Condition()"

    def as_dict(self):
        result = {"semantic_object_type": SEMANTIC_OBJECT_TYPE, "type": self.condition_type}
        return super().as_dict() | result


class HasValue(Condition):
    def __init__(self, predicate, type_=HAS_VALUE):
        super().__init__(type_)
        self._predicate = predicate

    @property
    def predicate(self):
        return self._predicate

    def is_true_given_proposition_set(self, proposition_set):
        for proposition in proposition_set:
            try:
                if proposition.get_predicate() == self.predicate:
                    return True
            except AttributeError:
                pass
        return False

    def __eq__(self, other):
        try:
            equality = super().__eq__(other) and self.predicate == other.predicate
            return equality
        except AttributeError:
            return False

    def __repr__(self):
        return "{}({})".format(self.condition_type, self.predicate)


class HasSharedValue(HasValue):
    def __init__(self, predicate):
        super().__init__(predicate, HAS_SHARED_VALUE)

    def is_true_given_tis(self, tis):
        return super().is_true_given_proposition_set(tis.shared.com)


class HasPrivateValue(HasValue):
    def __init__(self, predicate):
        super().__init__(predicate, HAS_PRIVATE_VALUE)

    def is_true_given_tis(self, tis):
        return super().is_true_given_proposition_set(tis.private.bel)


class HasSharedOrPrivateValue(HasValue):
    def __init__(self, predicate):
        super().__init__(predicate, HAS_SHARED_OR_PRIVATE_VALUE)

    def is_true_given_tis(self, tis):
        return super().is_true_given_proposition_set(tis.private.bel
                                                     ) or super().is_true_given_proposition_set(tis.shared.com)


class IsTrue(Condition):
    def __init__(self, proposition, type_=IS_TRUE):
        super().__init__(type_)
        self._proposition = proposition

    @property
    def proposition(self):
        return self._proposition

    def is_true_given_proposition_set(self, proposition_set):
        return self.proposition in proposition_set

    def __eq__(self, other):
        try:
            return super().__eq__(other) and self.proposition == other.proposition
        except AttributeError:
            return False

    def __repr__(self):
        return "{}({})".format(self.condition_type, self.proposition)


class IsSharedCommitment(IsTrue):
    def __init__(self, proposition):
        super().__init__(proposition, IS_SHARED_COMMITMENT)

    def is_true_given_tis(self, tis):
        return super().is_true_given_proposition_set(tis.shared.com)


class IsPrivateBelief(IsTrue):
    def __init__(self, proposition):
        super().__init__(proposition, IS_PRIVATE_BELIEF)

    def is_true_given_tis(self, tis):
        return super().is_true_given_proposition_set(tis.private.bel)


class IsPrivateBeliefOrSharedCommitment(IsTrue):
    def __init__(self, proposition):
        super().__init__(proposition, IS_PRIVATE_BELIEF_OR_SHARED_COMMITMENT)

    def is_true_given_tis(self, tis):
        return super().is_true_given_proposition_set(tis.private.bel
                                                     ) or super().is_true_given_proposition_set(tis.shared.com)


class QueryHasMoreItems(Condition):
    def __init__(self, query, type_=QUERY_HAS_MORE_ITEMS):
        super().__init__(type_)
        self._query = query

    @property
    def query(self):
        return self._query

    def is_true_given_tis(self, tis):
        no_more_elements = tis.iterator_has_no_more_elements(self.query)
        reached_limit = tis.iterator_has_reached_limit(self.query)
        result = not (no_more_elements or reached_limit)
        return result

    def __eq__(self, other):
        try:
            equality = super().__eq__(other) and self.query == other.query
            return equality
        except AttributeError:
            return False

    def __repr__(self):
        return "{}({})".format(self.condition_type, self.query)



from copy import copy

from tala.utils.as_semantic_expression import AsSemanticExpressionMixin


def json_semantic_expression_of(object_):
    return {"semantic_expression": object_.as_semantic_expression()}


def convert_to_json(object_, verbose=True):
    if object_ is None:
        return None
    if object_ is True or object_ is False:
        return object_
    # if isinstance(object_, float) or isinstance(object_, int):
    #     return object_
    if isinstance(object_, list):
        return [convert_to_json(element, verbose) for element in object_]
    if isinstance(object_, set):
        return {"set": [convert_to_json(element, verbose) for element in object_]}
    if isinstance(object_, dict):
        return {str(key): convert_to_json(value, verbose) for key, value in list(object_.items())}
    if not verbose and isinstance(object_, AsSemanticExpressionMixin):
        return json_semantic_expression_of(object_)
    if isinstance(object_, AsJSONMixin):
        dict_ = object_.as_dict()
        json = convert_to_json(dict_, verbose)
        if isinstance(object_, AsSemanticExpressionMixin):
            json.update(json_semantic_expression_of(object_))
        return json
    return str(object_)


class AsJSONMixin(object):
    @property
    def can_convert_to_json(self):
        return True

    def as_json(self):
        return convert_to_json(self, verbose=True)

    def as_compact_json(self):
        return convert_to_json(self, verbose=False)

    def as_dict(self):
        return copy(self.__dict__)



from tala.model.polarity import Polarity


class HttpFormatter(object):
    def __init__(self, ddd_component_manager):
        self._ddd_component_manager = ddd_component_manager

    def facts_to_json_object(self, facts, session):
        return {
            proposition.getPredicate().get_name(): self.fact_to_json_object(proposition, session)
            for proposition in facts
            if proposition.is_predicate_proposition() and proposition.get_polarity() == Polarity.POS
        }

    def fact_to_json_object(self, proposition, session):
        def get_grammar_entry(value):
            if "entities" in session:
                for entity in session["entities"]:
                    if entity["name"] == value:
                        return entity["natural_language_form"]
            return None

        if proposition is None:
            return None
        individual = proposition.getArgument()
        if individual is None:
            return None
        value = individual.getValue()
        value_as_json = individual.value_as_json_object()
        grammar_entry = get_grammar_entry(value)
        resulting_dict = {
            "sort": proposition.getPredicate().getSort().get_name(),
            "grammar_entry": grammar_entry,
            "perception_confidence":
                proposition.confidence_estimates.perception_confidence if proposition.confidence_estimates else None,
            "understanding_confidence":
                proposition.confidence_estimates.understanding_confidence if proposition.confidence_estimates else None,
            "weighted_confidence":
                proposition.confidence_estimates.weighted_confidence if proposition.confidence_estimates else None,
            "weighted_understanding_confidence":
                proposition.confidence_estimates.weighted_understanding_confidence
                if proposition.confidence_estimates else None
        }  # yapf: disable
        resulting_dict.update(value_as_json)
        return resulting_dict



import unittest

from mock import Mock

from tala.ddd.services.service_interface import ServiceInterface
from tala.model.ddd import DDD
from tala.model.domain import Domain
from tala.model.lambda_abstraction import LambdaAbstractedPredicateProposition, LambdaAbstractedGoalProposition
from tala.model.ontology import Ontology
from tala.model.polarity import Polarity
from tala.model.predicate import Predicate
from tala.model.proposition import PredicateProposition
from tala.model.question import WhQuestion
from tala.model.sort import CustomSort, RealSort, IntegerSort, BooleanSort, StringSort
from tala.testing.utils import EqualityAssertionTestCaseMixin


class LibTestCase(unittest.TestCase, EqualityAssertionTestCaseMixin):
    DDD_NAME = "mockup_ddd"

    def setUpLibTestCase(self):
        self.ontology_name = "mockup_ontology"
        self._city_sort = CustomSort(self.ontology_name, "city", dynamic=True)
        self.ontology = self._create_ontology()

        self.empty_ontology = Ontology("empty_ontology", {}, {}, {}, set())

        self.domain_name = "mockup_domain"

        self._create_semantic_objects()
        self.ddd = self._create_and_add_ddd()

    def _create_ontology(self):
        sorts = {
            self._city_sort,
            CustomSort(self.ontology_name, "city_type"),
            CustomSort(self.ontology_name, "ticket_type"),
            CustomSort(self.ontology_name, "passenger_type"),
        }
        predicates = {
            self._create_predicate("dest_city", self._city_sort),
            self._create_predicate(
                "dest_city_type", sort=CustomSort(self.ontology_name, "city_type"), feature_of_name="dest_city"
            ),
            self._create_predicate("dept_city", self._city_sort),
            self._create_predicate("price", RealSort()),
            self._create_predicate("number_of_passengers", IntegerSort()),
            self._create_predicate(
                "passenger_type_to_add", sort=CustomSort(self.ontology_name, "passenger_type"), multiple_instances=True
            ),
            self._create_predicate("available_ticket_type", CustomSort(self.ontology_name, "ticket_type")),
            self._create_predicate("available_city", self._city_sort),
            self._create_predicate("need_visa", BooleanSort()),
            self._create_predicate("comment_message", StringSort()),
        }
        individuals = {
            "paris": self._city_sort,
            "london": self._city_sort,
        }
        actions = {"top", "buy"}
        return Ontology(self.ontology_name, sorts, predicates, individuals, actions)

    def _create_predicate(self, *args, **kwargs):
        return Predicate(self.ontology_name, *args, **kwargs)

    def _create_semantic_objects(self):
        self.sort_city = self.ontology.get_sort("city")
        self.sort_city_type = self.ontology.get_sort("city_type")

        self.predicate_dest_city = self.ontology.get_predicate("dest_city")
        self.predicate_dest_city_type = self.ontology.get_predicate("dest_city_type")
        self.predicate_dept_city = self.ontology.get_predicate("dept_city")
        self.predicate_price = self.ontology.get_predicate("price")
        self.predicate_available_ticket_type = self.ontology.get_predicate("available_ticket_type")
        self.predicate_available_city = self.ontology.get_predicate("available_city")
        self.predicate_number_of_passengers = self.ontology.get_predicate("number_of_passengers")
        self.predicate_need_visa = self.ontology.get_predicate("need_visa")

        self.individual_paris = self.ontology.create_individual("paris")
        self.individual_london = self.ontology.create_individual("london")
        self.individual_not_paris = self.ontology.create_negative_individual("paris")

        self.real_individual = self.ontology.create_individual(1234.0)

        self.proposition_dest_city_paris = PredicateProposition(self.predicate_dest_city, self.individual_paris)
        self.proposition_dest_city_london = PredicateProposition(self.predicate_dest_city, self.individual_london)
        self.proposition_dept_city_paris = PredicateProposition(self.predicate_dept_city, self.individual_paris)
        self.proposition_dept_city_london = PredicateProposition(self.predicate_dept_city, self.individual_london)
        self.proposition_not_dest_city_paris = PredicateProposition(
            self.predicate_dest_city, self.individual_paris, Polarity.NEG
        )
        self.price_proposition = PredicateProposition(self.predicate_price, self.real_individual)

        self.lambda_abstracted_price_prop = LambdaAbstractedPredicateProposition(
            self.predicate_price, self.ontology_name
        )
        self.lambda_abstracted_dest_city_prop = LambdaAbstractedPredicateProposition(
            self.predicate_dest_city, self.ontology_name
        )
        self.lambda_abstracted_dept_city_prop = LambdaAbstractedPredicateProposition(
            self.predicate_dept_city, self.ontology_name
        )
        self.lambda_abstracted_ticket_type_prop = LambdaAbstractedPredicateProposition(
            self.predicate_available_ticket_type, self.ontology_name
        )
        self.lambda_abstracted_available_city_prop = LambdaAbstractedPredicateProposition(
            self.predicate_available_city, self.ontology_name
        )
        self.lambda_abstracted_number_of_passengers_prop = LambdaAbstractedPredicateProposition(
            self.predicate_number_of_passengers, self.ontology_name
        )

        self.price_question = WhQuestion(self.lambda_abstracted_price_prop)
        self.dest_city_question = WhQuestion(self.lambda_abstracted_dest_city_prop)
        self.dept_city_question = WhQuestion(self.lambda_abstracted_dept_city_prop)
        self.action_question = WhQuestion(LambdaAbstractedGoalProposition())
        self.available_ticket_type_question = WhQuestion(self.lambda_abstracted_ticket_type_prop)
        self.available_city_question = WhQuestion(self.lambda_abstracted_available_city_prop)
        self.number_of_passengers_question = WhQuestion(self.lambda_abstracted_number_of_passengers_prop)

        self.lambda_abstracted_goal_prop = LambdaAbstractedGoalProposition()

        self.buy_action = self.ontology.create_action("buy")
        self.top_action = self.ontology.create_action("top")

    def _create_and_add_ddd(self):
        return DDD(
            self.DDD_NAME,
            self.ontology,
            Domain(self.DDD_NAME, "mockup_domain", self.ontology),
            service_interface=Mock(spec=ServiceInterface),
        )

    def then_result_is(self, expected_result):
        self.assertEqual(expected_result, self._actual_result)

    def when_call(self, callable_):
        self._actual_result = callable_()



from mock import patch
import unittest

import tala.testing.endurance.named_test
from tala.testing.endurance.named_test import EnduranceInteractionTest, MissingTurnsException
from tala.testing.interactions.named_test import InteractionTest
from tala.testing.interactions.turn import RecognitionHypothesesTurn, SystemUtteranceTurn


class EnduranceInteractionTestTests(unittest.TestCase):
    def setUp(self):
        self._interaction_tests = []
        self._duration = None
        self._endurance_test = None
        self._actual_turn = None
        self._turns_iterator = None
        self._mock_time = None
        self._mock_random = None

    def test_first_turn_is_of_type_system(self):
        self._given_mocked_tests([
            self._create_interaction_test_with_turns([
                SystemUtteranceTurn("mock system utterance", 1),
                RecognitionHypothesesTurn(["mock user utterance"], 2)
            ])])
        self._given_duration(0)
        self._given_created_endurance_test()
        self._given_fetched_turns_iterator()
        self._when_fetching_the_next_turn()
        self._then_it_is_recognition_hypotheses_turn()

    @staticmethod
    def _create_interaction_test_with_turns(turns):
        mocked_test = InteractionTest("mock_file", "mock_name", turns)
        return mocked_test

    def _given_mocked_tests(self, tests):
        self._interaction_tests = tests

    def _given_duration(self, duration):
        self._duration = duration

    def _given_created_endurance_test(self):
        self._endurance_test = EnduranceInteractionTest(self._interaction_tests, self._duration)

    def _given_fetched_turns_iterator(self):
        self._turns_iterator = self._endurance_test.turns

    def _when_fetching_the_next_turn(self):
        self._actual_turn = next(self._turns_iterator)

    def _then_it_is_recognition_hypotheses_turn(self):
        self.assertTrue(self._actual_turn.is_recognition_hypotheses_turn)

    def test_only_turn_is_of_type_system(self):
        self._given_mocked_tests([
            self._create_interaction_test_with_turns([
                SystemUtteranceTurn("mock system utterance", 1)
            ])])
        self._given_duration(0)
        self._given_created_endurance_test()
        self._given_fetched_turns_iterator()
        self._when_fetching_the_next_turn_then_exception_is_raised_matching(
            MissingTurnsException, "Expected turns but found none in test 'mock_name' of 'mock_file'")

    def _when_fetching_the_next_turn_then_exception_is_raised_matching(self, ExceptionClass, message_regex):
        with self.assertRaisesRegex(ExceptionClass, message_regex):
            next(self._turns_iterator)

    def _given_fetched_the_next_turn(self):
        next(self._turns_iterator)

    @patch("%s.time" % tala.testing.endurance.named_test.__name__)
    def test_turns_with_positive_duration(self, mock_time):
        self._given_mock_time(mock_time)
        self._given_mock_time_is(0)
        self._given_mocked_tests([
            self._create_interaction_test_with_turns([
                RecognitionHypothesesTurn(["mock user utterance"], 1)
            ])])
        self._given_duration(1)
        self._given_created_endurance_test()
        self._given_fetched_turns_iterator()
        self._given_fetched_the_next_turn()
        self._given_mock_time_is(2)
        self._when_fetching_the_next_turn_then_stop_iteration_is_raised()

    def _when_fetching_the_next_turn_then_stop_iteration_is_raised(self):
        with self.assertRaises(StopIteration):
            next(self._turns_iterator)

    def _given_mock_time(self, mock_time):
        self._mock_time = mock_time

    def _given_mock_time_is(self, time_now):
        self._mock_time.time.return_value = time_now

    def test_stop(self):
        self._given_mocked_tests([
            self._create_interaction_test_with_turns([
                RecognitionHypothesesTurn(["mock user utterance"], 1)
            ])])
        self._given_duration(0)
        self._given_created_endurance_test()
        self._given_fetched_turns_iterator()
        self._given_fetched_the_next_turn()
        self._given_called_stop()
        self._when_fetching_the_next_turn_then_stop_iteration_is_raised()

    def _given_called_stop(self):
        self._endurance_test.stop()

    def test_turns_are_generated_over_and_over_again(self):
        self._given_mocked_tests([
            self._create_interaction_test_with_turns([
                RecognitionHypothesesTurn(["mock user utterance"], 1)
            ])])
        self._given_duration(0)
        self._given_created_endurance_test()
        self._given_fetched_turns_iterator()
        self._given_fetched_the_next_turn()  # U>
        self._given_fetched_the_next_turn()  # PASSIVITY
        self._given_fetched_the_next_turn()  # S> *
        self._when_fetching_the_next_turn()  # U> again
        self._then_it_is_recognition_hypotheses_turn()

    @patch("%s.random" % tala.testing.endurance.named_test.__name__)
    def test_turns_are_generated_randomly(self, mock_random):
        self._given_mock_random(mock_random)
        self._given_mock_random_choice_returns(
            self._create_interaction_test_with_turns([
                RecognitionHypothesesTurn(["mock user utterance"], 1)
            ]))
        self._given_mocked_tests([
            self._create_interaction_test_with_turns([
                RecognitionHypothesesTurn(["mock user utterance"], 1)
            ])])
        self._given_duration(0)
        self._given_created_endurance_test()
        self._given_fetched_turns_iterator()
        self._when_fetching_the_next_turn()
        self._then_next_interaction_test_was_randomly_selected()

    def _given_mock_random(self, mock_random):
        self._mock_random = mock_random

    def _given_mock_random_choice_returns(self, test):
        self._mock_random.choice.return_value = test

    def _then_next_interaction_test_was_randomly_selected(self):
        self.assertIs(self._actual_turn, self._mock_random.choice.return_value.turns[0])



import unittest

from tala.model.openqueue import OpenQueue, OpenQueueError


class MockElement:
    def __init__(self, name):
        self._name = name

    def __str__(self):
        return "MockElement(%r)" % self._name


class OpenQueueTests(unittest.TestCase):
    def setUp(self):
        self.queue = OpenQueue()

    def test_first(self):
        self._given_an_element_enqueued("a")
        self._when_an_element_enqueued("b")
        self._first_returns("a")

    def _given_an_element_enqueued(self, elem):
        self.queue.enqueue(elem)

    def _when_an_element_enqueued(self, elem):
        self.queue.enqueue(elem)

    def _first_returns(self, element):
        actual_element = self.queue.first()
        self.assertEqual(element, actual_element)

    def test_is_first(self):
        self._given_an_element_enqueued("a")
        self._is_first("a")

    def _is_first(self, element):
        self.assertTrue(self.queue.is_first(element))

    def test_is_not_first_on_empty_queue(self):
        self._is_not_first("a")

    def _is_not_first(self, element):
        self.assertFalse(self.queue.is_first(element))

    def test_first_on_empty_queue_raises_exception(self):
        with self.assertRaises(OpenQueueError):
            self.queue.first()

    def test_last(self):
        self._given_an_element_enqueued("a")
        self._when_an_element_enqueued("b")
        self._last_returns("b")

    def _last_returns(self, element):
        actual_element = self.queue.last()
        self.assertEqual(element, actual_element)

    def test_size_0(self):
        self.assertEqual(0, len(self.queue))

    def test_size_1(self):
        self._given_an_element_enqueued("a")
        self.assertEqual(1, len(self.queue))

    def test_dequeue_returns_right_element(self):
        self._given_an_element_enqueued("a")
        self._dequeue_returns("a")

    def _dequeue_returns(self, element):
        actual_element = self.queue.dequeue()
        self.assertEqual(element, actual_element)

    def test_dequeue_deletes_element(self):
        self._given_an_element_enqueued("a")
        self._when_dequeue()
        self._queue_is_empty()

    def _when_dequeue(self):
        self.queue.dequeue()

    def _queue_is_empty(self):
        self.assertTrue(len(self.queue) == 0)

    def test_dequeue_on_empty_queue_raises_exception(self):
        with self.assertRaises(OpenQueueError):
            self.queue.dequeue()

    def test_delete(self):
        self._given_an_element_enqueued("a")
        self._given_an_element_enqueued("b")
        self._when_an_element_is_deleted("b")
        self._element_is_not_member("b")

    def _when_an_element_is_deleted(self, element):
        self.queue.remove(element)

    def _element_is_not_member(self, non_member):
        self.assertFalse(non_member in self.queue)

    def test_string_representation(self):
        self.queue.enqueue(MockElement("first"))
        self.queue.enqueue(MockElement("second"))
        self.assertEqual("OpenQueue(['#', MockElement('first'), MockElement('second')])", str(self.queue))

    def test_create_from_iterable(self):
        self.queue = OpenQueue(["a", "b", "c"])
        self.assertEqual("OpenQueue(['a', 'b', 'c', '#'])", str(self.queue))

    def test_shift(self):
        self.queue = OpenQueue(["a", "b", "c"])
        self.queue.init_shift()
        self.queue.shift()
        self.assertEqual("OpenQueue(['b', 'c', '#', 'a'])", str(self.queue))

    def test_fully_shifted(self):
        self.queue = OpenQueue(["a", "b", "c"])
        self.queue.init_shift()
        self.queue.shift()
        self.queue.shift()
        self.queue.shift()
        self.assertTrue(self.queue.fully_shifted())

    def test_dequeue_and_shift(self):
        self.queue = OpenQueue(["a", "b", "c"])
        self.queue.dequeue()
        self.queue.init_shift()
        self.queue.shift()
        self.assertEqual("OpenQueue(['c', '#', 'b'])", str(self.queue))

    def test_clear(self):
        self._given_an_element_enqueued("element")
        self._when_clear_is_called()
        self._object_is_emptied()

    def _when_clear_is_called(self):
        self.queue.clear()

    def _object_is_emptied(self):
        self.assertTrue(self.queue.is_empty())

    def test_empty_queues_are_equal(self):
        first_queue = OpenQueue()
        second_queue = OpenQueue()
        self.assertEqual(first_queue, second_queue)

    def test_empty_queue_not_equals_none(self):
        empty_queue = OpenQueue()
        self.assertNotEqual(None, empty_queue)

    def test_single_element_queues_are_equal(self):
        first_queue = OpenQueue(["a"])
        second_queue = OpenQueue(["a"])
        self.assertEqual(first_queue, second_queue)

    def test_single_element_queue_equal_to_shifted_single_element_queue(self):
        first_queue = OpenQueue(["a"])
        second_queue = OpenQueue(["a"])
        second_queue.shift()
        self.assertEqual(first_queue, second_queue)

    def test_unshifted_multi_element_queues_are_equal(self):
        first_queue = OpenQueue(["a", "b", "c"])
        second_queue = OpenQueue(["a", "b", "c"])
        self.assertEqual(first_queue, second_queue)

    def test_queue_not_equal_to_shifted_queue_1(self):
        first_queue = OpenQueue(["a", "b", "c"])
        second_queue = OpenQueue(["a", "b", "c"])
        second_queue.shift()
        self.assertNotEqual(first_queue, second_queue)

    def test_queue_not_equal_to_shifted_queue_2(self):
        first_queue = OpenQueue(["a", "b", "c"])
        second_queue = OpenQueue(["c", "a", "b"])
        second_queue.shift()
        self.assertNotEqual(first_queue, second_queue)

    def test_enqueue_first(self):
        queue = OpenQueue(["b"])
        queue.enqueue_first("a")
        self.assertEqual("a", queue.first())

    def test_remove_if_exists_for_existing_element(self):
        queue = OpenQueue(["first", "second"])
        queue.remove_if_exists("first")
        expected_result = OpenQueue(["second"])
        self.assertEqual(expected_result, queue)

    def test_remove_if_exists_for_non_existing_element_has_no_effect(self):
        queue = OpenQueue(["first", "second"])
        queue.remove_if_exists("third")
        expected_result = OpenQueue(["first", "second"])
        self.assertEqual(expected_result, queue)

    def test_set_property(self):
        self._given_an_element_enqueued("a")
        self._when_an_element_enqueued("a")
        self._first_returns("a")
        self._len_is(1)

    def _len_is(self, length):
        self.assertEqual(length, len(self.queue))

    def test_cancel_shift(self):
        queue = OpenQueue(["a", "b", "c"])
        queue.init_shift()
        queue.shift()
        queue.remove("b")
        queue.cancel_shift()
        expected_queue = OpenQueue(["a", "c"])
        self.assertEqual(expected_queue, queue)

    def test_cancel_shift_raises_exception_if_init_shift_not_called(self):
        queue = OpenQueue([])
        with self.assertRaises(OpenQueueError):
            queue.cancel_shift()

    def test_getitem(self):
        queue = OpenQueue(["a", "b", "c"])
        self.assertEqual("a", queue[0])



import pytest

from .console_script_mixin import ConsoleScriptTestMixin


@pytest.mark.skip("Issues with prompt toolkit not having a stdout on vt100")
class TestInteractIntegration(ConsoleScriptTestMixin):
    def test_interacting_with_unknown_environment(self):
        self._given_created_deployments_config()
        self._run_tala_with(["tala", "interact", "my-made-up-environment"])
        self._then_stdout_matches(
            r"Expected a URL or one of the known environments \['dev'\] but got 'my-made-up-environment'"
        )

    def _given_created_deployments_config(self):
        self._run_tala_with(["create-deployments-config"])



from tala.utils.as_json import AsJSONMixin


class SemanticObject(AsJSONMixin):
    def is_yes(self):
        return False

    def is_no(self):
        return False

    def is_individual(self):
        return False

    def is_proposition(self):
        return False

    def is_question(self):
        return False

    def is_action(self):
        return False

    def is_goal(self):
        return False

    def is_plan_item(self):
        return False

    def is_predicate_proposition(self):
        return False

    def is_proposition_set(self):
        return False

    def is_lambda_abstracted_predicate_proposition(self):
        return False

    def is_lambda_abstracted_goal_proposition(self):
        return False

    def is_lambda_abstracted_domain_proposition(self):
        return False

    def is_move(self):
        return False

    def is_predicate(self):
        return False

    @property
    def ontology_name(self):
        message = "%s derives from %s and does not have an ontology name. Use %s instead." % (
            self, SemanticObject.__name__, OntologySpecificSemanticObject.__name__
        )
        raise NotImplementedError(message)

    def is_ontology_specific(self):
        return False

    def has_semantic_content(self):
        return False

    def __repr__(self):
        return repr(self.__dict__)


class OntologySpecificSemanticObject(SemanticObject):
    def __init__(self, ontology_name):
        self._ontology_name = ontology_name

    @property
    def ontology_name(self):
        return self._ontology_name

    def is_ontology_specific(self):
        return True


class SemanticObjectWithContent(SemanticObject):
    def __init__(self, content):
        self._content = content

    @property
    def ontology_name(self):
        return self._content.ontology_name

    def is_ontology_specific(self):
        return self._content.is_ontology_specific()

    def has_semantic_content(self):
        return True



import unittest


class TestSuite(unittest.TestSuite):
    def __init__(self, filename):
        super().__init__()
        self._filename = filename

    @property
    def filename(self):
        return self._filename



from tala.testing.interactions.suite import TestSuite
from tala.testing.interactions.testcase import InteractionTestingTestCase


class InteractionTestingLoader(object):
    def __init__(self, url):
        self._url = url

    def load_interaction_tests(self, interaction_testing_file, selected_tests=None):
        suite = TestSuite(interaction_testing_file.filename)
        tests = interaction_testing_file.tests
        if selected_tests:
            tests = interaction_testing_file.filter_tests_by_name(selected_tests)
        for test in tests:
            test_case = InteractionTestingTestCase(test, self._url)
            suite.addTest(test_case)
        return suite



from tala.model.goal import Goal, ResolveGoal, PerformGoal
from tala.model.speaker import Speaker
from tala.testing.lib_test_case import LibTestCase


class GoalTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_equality(self):
        goal = Goal("type 1", Speaker.SYS)
        identical_goal = Goal("type 1", Speaker.SYS)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(identical_goal, goal)

    def test_is_goal_false_by_default(self):
        non_goal_object = self.top_action
        self.assertFalse(non_goal_object.is_goal())


class ResolveGoalTest(LibTestCase):
    def test_equality(self):
        goal = ResolveGoal(self.price_question, Speaker.SYS)
        identical_goal = ResolveGoal(self.price_question, Speaker.SYS)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(identical_goal, goal)

    def test_inequality_due_to_question(self):
        goal1 = ResolveGoal(self.dest_city_question, Speaker.SYS)
        goal2 = ResolveGoal(self.price_question, Speaker.SYS)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(goal1, goal2)

    def test_inequality_due_to_target(self):
        goal1 = ResolveGoal(self.dest_city_question, Speaker.USR)
        goal2 = ResolveGoal(self.dest_city_question, Speaker.SYS)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(goal1, goal2)

    def test_hashable(self):
        goal = ResolveGoal(self.price_question, Speaker.SYS)
        {goal}

    def test_is_goal(self):
        goal = ResolveGoal(self.price_question, Speaker.SYS)
        self.assertTrue(goal.is_goal())

    def test_str_with_target_sys(self):
        goal = ResolveGoal(self.price_question, Speaker.SYS)
        self.assertEqual("resolve(?X.price(X))", str(goal))

    def test_str_with_target_user(self):
        goal = ResolveGoal(self.price_question, Speaker.USR)
        self.assertEqual("resolve_user(?X.price(X))", str(goal))

    def test_str_with_background(self):
        goal = ResolveGoal(self.price_question, Speaker.SYS)
        goal.set_background(self.predicate_dest_city)
        self.assertEqual("resolve(?X.price(X), dest_city)", str(goal))

    def setUp(self):
        self.setUpLibTestCase()


class PerformGoalTest(LibTestCase):
    def test_equality(self):
        goal = PerformGoal(self.top_action)
        identical_goal = PerformGoal(self.top_action)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(identical_goal, goal)

    def test_inequality_due_to_action(self):
        goal1 = PerformGoal(self.buy_action)
        goal2 = PerformGoal(self.top_action)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(goal1, goal2)

    def test_hashable(self):
        goal = PerformGoal(self.top_action)
        {goal}

    def test_is_goal(self):
        goal = PerformGoal(self.top_action)
        self.assertTrue(goal.is_goal())

    def setUp(self):
        self.setUpLibTestCase()



import re
import subprocess

from .console_script_mixin import ConsoleScriptTestMixin


class TestVersionIntegration(ConsoleScriptTestMixin):
    def test_version(self):
        self._when_checking_tala_version()
        self._then_result_is_a_version_number()

    def _when_checking_tala_version(self):
        process = subprocess.Popen(["tala", "version"], stdout=subprocess.PIPE, text=True)
        stdout, _ = process.communicate()
        self._result = stdout

    def _then_result_is_a_version_number(self):
        base_version = r"[0-9]+\.[0-9]+(\.[0-9]+)*"
        dev_version_suffix = r"\.dev[0-9]+\+[a-z0-9]+"
        dirty_version_suffix = r"\.d[0-9]{8}"

        release_version_regexp = base_version
        dev_version_regexp = base_version + dev_version_suffix
        dirty_version_regexp = base_version + dev_version_suffix + dirty_version_suffix

        is_version_regexp = r"(?:%s|%s|%s)" % (release_version_regexp, dev_version_regexp, dirty_version_regexp)

        assert re.search(is_version_regexp, self._result) is not None



import unittest

from tala.event import Event
from tala.model.set import Set


class EventTests(unittest.TestCase):
    def test_input_event(self):
        event = Event(Event.PASSIVITY, "hello")
        self.assertEqual(Event.PASSIVITY, event.get_type())
        self.assertEqual("hello", event.get_content())

    def test_event_equality_for_input(self):
        event = Event(Event.PASSIVITY, "hello")
        identical_event = Event(Event.PASSIVITY, "hello")
        self.assertEqual(event, identical_event)

    def test_event_equality_for_interpretation(self):
        moves = Set(["move1", "move2"])
        event = Event(Event.INTERPRETATION, moves)
        identical_moves = Set(["move1", "move2"])
        identical_event = Event(Event.INTERPRETATION, identical_moves)
        self.assertEqual(event, identical_event)
        self.assertEqual(identical_event, event)
        self.assertFalse(event != identical_event)
        self.assertFalse(identical_event != event)

    def test_events_inequal_due_to_content(self):
        event = Event(Event.PASSIVITY, "hello")
        non_identical_event = Event(Event.PASSIVITY, "goodbye")
        self.assertNotEqual(event, non_identical_event)
        self.assertNotEqual(non_identical_event, event)

    def test_events_inequal_due_to_sender(self):
        event = Event(Event.PASSIVITY, "hello", sender="sender1")
        non_identical_event = Event(Event.PASSIVITY, "hello", sender="sender2")
        self.assertNotEqual(event, non_identical_event)
        self.assertNotEqual(non_identical_event, event)

    def test_events_inequal_due_to_reason(self):
        event = Event(Event.PASSIVITY, "hello", reason="reason1")
        non_identical_event = Event(Event.PASSIVITY, "hello", reason="reason2")
        self.assertNotEqual(event, non_identical_event)
        self.assertNotEqual(non_identical_event, event)

    def test_event_not_equals_none(self):
        event = Event(Event.PASSIVITY, "hello")
        self.assertNotEqual(None, event)

    def test_string_representation_for_non_empty_event(self):
        event = Event(Event.PASSIVITY, "hello")
        self.assertEqual("Event(PASSIVITY, 'hello', sender=None, reason=None)", str(event))

    def test_string_representation_for_empty_event(self):
        event = Event(Event.STOP)
        self.assertEqual("Event(STOP, None, sender=None, reason=None)", str(event))

    def test_is_internal(self):
        self.assertTrue(Event(Event.STOP).is_internal())
        self.assertFalse(Event(Event.PASSIVITY).is_internal())



from copy import copy


class Observable(object):
    def __init__(self):
        self._observers = set()

    def add_observer(self, observer):
        self._observers.add(observer)

    def remove_observer(self, observer):
        self._observers.remove(observer)

    def set(self, value):
        for observer in copy(self._observers):
            observer.update(value)


class Observer(object):
    def update(self, value):
        raise NotImplementedError()



from tala.model.set import Set
from tala.utils.as_json import AsJSONMixin
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin
from tala.ddd.json_parser import NonCheckingJSONParser


class Commitments(AsSemanticExpressionMixin, AsJSONMixin):
    def __init__(self, content=None, meta_com=None, pcom=None, meta_pcom=None, previous_commitment_lifespan=30):
        if content is None:
            content = []
        if pcom is None:
            pcom = []
        if meta_com is None:
            meta_com = {}
        if meta_pcom is None:
            meta_pcom = {}
        self._backing_set = Set([item for item in content])
        self._meta_pcom = {k: v for k, v in meta_pcom.items()}
        self._meta_com = {k: v for k, v in meta_com.items()}
        self._pcom = Set([item for item in pcom])
        self._previous_commitment_lifespan = previous_commitment_lifespan

    @classmethod
    def create_from_json(cls, input):
        parser = NonCheckingJSONParser()
        commitments = [parser.parse(commitment) for commitment in input["commitments_set"]["set"]]
        lifespan = int(input["commitment_lifespan"])
        meta_com = {parser.parse(elem["key"]): (elem["value"][0], elem["value"][1]) for elem in input["meta_com"]}
        pcom = [parser.parse(commitment) for commitment in input["commitments_set"]["set"]]
        meta_pcom = {
            parser.parse(elem["key"]): (elem["value"][0], int(elem["value"][1]))
            for elem in input["meta_pcom"]
        }
        return cls(commitments, meta_com, pcom, meta_pcom, lifespan)

    @property
    def meta_com(self):
        return self._meta_com

    @property
    def pcom(self):
        return self._pcom

    @property
    def meta_pcom(self):
        return self._meta_pcom

    def as_dict(self):
        return {
            "commitments_set": self._backing_set,
            "meta_com": [{
                "key": k,
                "value": [v[0], v[1]]
            } for k, v in self._meta_com.items()],
            "pcom": self._pcom,
            "meta_pcom": [{
                "key": k,
                "value": [v[0], v[1]]
            } for k, v in self._meta_pcom.items()],
            "commitment_lifespan": self._previous_commitment_lifespan
        }

    def union(self, other):
        return self._backing_set.union(other)

    def add(self, proposition, turn_number=-1, topmost_goal=None):
        self._remove_incompatible_commitments(proposition)
        self._backing_set.add(proposition)
        self._meta_com[proposition] = (topmost_goal, turn_number)

    def _remove_incompatible_commitments(self, proposition):
        to_remove = []

        for commitment in self:
            if proposition.is_incompatible_with(commitment):
                to_remove.append(commitment)

        for commitment_to_remove in to_remove:
            self.remove(commitment_to_remove)

    def __iter__(self):
        return self._backing_set.__iter__()

    def remove(self, proposition):
        if proposition in self._meta_com:
            self._meta_pcom[proposition] = self._meta_com[proposition]
        self._backing_set.remove(proposition)
        self.pcom.add(proposition)

    def remove_if_exists(self, proposition):
        if proposition in self._backing_set:
            self.remove(proposition)

    def __str__(self):
        return str(self._backing_set)

    def __repr__(self):
        return repr(self._backing_set)

    def __eq__(self, other):
        return self._backing_set.__eq__(other)

    def is_subset_of(self, other):
        return self._backing_set.is_subset_of(other)

    def __len__(self):
        return len(self._backing_set)

    def should_purge_previous_commitments(self, current_turn_number):
        for fact in self.pcom:
            if fact in self._meta_pcom:
                (_, turn_number) = self._meta_pcom[fact]
                if int(turn_number) + self._previous_commitment_lifespan < current_turn_number:
                    return True
        return False

    def purge_previous_commitments(self, current_turn_number):
        to_purge = []

        for fact in self.pcom:
            if fact in self._meta_pcom:
                (_, turn_number) = self._meta_pcom[fact]
                if turn_number + self._previous_commitment_lifespan < current_turn_number:
                    to_purge.append(fact)

        for fact in to_purge:
            self.pcom.remove(fact)
            if fact in self._meta_pcom:
                del self._meta_pcom[fact]



import warnings

from tala.model.common import Modality
from tala.model.confidence_estimate import ConfidenceEstimates
from tala.model.speaker import Speaker
from tala.model.semantic_object import SemanticObject, OntologySpecificSemanticObject, SemanticObjectWithContent
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin
from tala.utils.equality import EqualityMixin
from tala.utils import float_comparison
from tala.utils.unicodify import unicodify


class MoveException(Exception):
    pass


class Move(SemanticObject, AsSemanticExpressionMixin, EqualityMixin):
    QUIT = "quit"
    GREET = "greet"
    THANK_YOU = "thanks"
    THANK_YOU_RESPONSE = "thank_you_response"
    INSULT_RESPONSE = "insult_response"
    INSULT = "insult"
    NO_MOVE = "no_move"
    MUTE = "mute"
    UNMUTE = "unmute"

    ASK = "ask"
    ANSWER = "answer"
    REQUEST = "request"
    REPORT = "report"
    PREREPORT = "prereport"

    def __init__(
        self,
        type,
        understanding_confidence=None,
        speaker=None,
        modality=None,
        weighted_understanding_confidence=None,
        utterance=None,
        ddd_name=None,
        perception_confidence=None
    ):
        SemanticObject.__init__(self)
        self._type = type
        self._speaker = None
        self._modality = None
        self._ddd_name = None
        self._background = None
        self._utterance = None
        self._confidence_estimates = ConfidenceEstimates()
        if (
            understanding_confidence is not None or speaker is not None or modality is not None or utterance is not None
            or ddd_name is not None or perception_confidence is not None
        ):
            self.set_realization_data(
                understanding_confidence=understanding_confidence,
                speaker=speaker,
                modality=modality,
                utterance=utterance,
                ddd_name=ddd_name,
                perception_confidence=perception_confidence
            )
            if weighted_understanding_confidence is not None:
                self.weighted_understanding_confidence = weighted_understanding_confidence

    def is_move(self):
        return True

    def get_type(self):
        warnings.warn("Move.get_type() is deprecated. Use Move.type_ instead.", DeprecationWarning, stacklevel=2)
        return self._type

    @property
    def type(self):
        warnings.warn("Move.type is deprecated. Use Move.type_ instead.", DeprecationWarning, stacklevel=2)
        return self._type

    @property
    def type_(self):
        return self._type

    @property
    def confidence_estimates(self):
        return self._confidence_estimates

    @property
    def perception_confidence(self):
        return self._confidence_estimates.perception_confidence

    @property
    def understanding_confidence(self):
        return self._confidence_estimates.understanding_confidence

    @property
    def weighted_understanding_confidence(self):
        return self._confidence_estimates.weighted_understanding_confidence

    @weighted_understanding_confidence.setter
    def weighted_understanding_confidence(self, confidence):
        self._confidence_estimates.weighted_understanding_confidence = confidence

    @property
    def confidence(self):
        return self._confidence_estimates.confidence

    @property
    def weighted_confidence(self):
        return self._confidence_estimates.weighted_confidence

    def get_speaker(self):
        return self._speaker

    def get_modality(self):
        return self._modality

    def get_utterance(self):
        return self._utterance

    def set_ddd_name(self, ddd_name):
        self._ddd_name = ddd_name

    def get_ddd_name(self):
        return self._ddd_name

    def set_realization_data(
        self,
        understanding_confidence=None,
        speaker=None,
        modality=None,
        utterance=None,
        ddd_name=None,
        perception_confidence=None
    ):
        self._confidence_estimates.set_realization_data(
            perception_confidence=perception_confidence, understanding_confidence=understanding_confidence
        )
        self._speaker = speaker
        self._modality = modality or Modality.SPEECH
        self._utterance = utterance
        self._ddd_name = ddd_name

    def is_realized(self):
        return self.understanding_confidence is not None or self._speaker is not None or self._modality is not None \
               or (self._speaker == Speaker.USR and self._ddd_name is not None)

    def _verify_realization_data(self, understanding_confidence=None, speaker=None, modality=None, ddd_name=None):
        if self.is_realized():
            raise MoveException("realization data already set")
        if speaker is None:
            raise MoveException("speaker must be supplied")
        if speaker == Speaker.SYS:
            if understanding_confidence is not None and understanding_confidence != 1.0:
                raise MoveException("understanding confidence below 1.0 not allowed for system moves")
        if speaker == Speaker.USR:
            if understanding_confidence is None:
                raise MoveException("understanding confidence must be supplied for user moves")
            if ddd_name is None:
                raise MoveException("ddd_name must be supplied")
        if modality not in Modality.MODALITIES:
            raise MoveException(f"unsupported modality: '{modality}'")

    def __hash__(self):
        return hash((
            self.__class__.__name__, self._type, self.perception_confidence, self.understanding_confidence,
            self.weighted_understanding_confidence, self._speaker, self._utterance, self._ddd_name
        ))

    def __str__(self):
        return self._get_semantic_expression(include_attributes=True)

    def _get_semantic_expression(self, include_attributes):
        string = "Move("
        string += self._type
        if include_attributes:
            string += self._build_string_from_attributes()
        string += ")"
        return string

    def _build_string_from_attributes(self):
        string = ""
        if self._ddd_name:
            string += f", ddd_name={self._ddd_name!r}"
        if self._speaker:
            string += f", speaker={self._speaker}"
        string += self._confidence_estimates.build_string_from_attributes()
        if self._modality:
            string += f", modality={self._modality}"
        if self._utterance:
            string += f", utterance={self._utterance!r}"
        return string

    def semantic_expression_without_realization_data(self):
        return self._get_semantic_expression(include_attributes=False)

    @staticmethod
    def _is_confidence_equal(this, other):
        if this is None:
            return other is None
        return float_comparison.isclose(this, other)

    def _are_understanding_confidences_equal(self, other):
        return self._is_confidence_equal(self.understanding_confidence,
                                         other.understanding_confidence) and self._is_confidence_equal(
                                             self.weighted_understanding_confidence,
                                             other.weighted_understanding_confidence
                                         )

    def _are_perception_confidences_equal(self, other):
        return self._is_confidence_equal(self.perception_confidence, other.perception_confidence)

    def class_internal_move_content_equals(self, other):
        return True

    def move_content_equals(self, other):
        try:
            return (self.type_ == other.type_ and self.class_internal_move_content_equals(other))
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def is_icm(self):
        return False

    def is_question_raising(self):
        return self.type_ == Move.ASK

    def is_turn_yielding(self):
        return self.type_ in [Move.ANSWER, Move.GREET]

    def to_rich_string(self):
        return f"{str(self)}:{self._speaker}:{self.understanding_confidence}"

    def uprank(self, amount):
        self._confidence_estimates.uprank(amount)

    def downrank(self, amount):
        self._confidence_estimates.downrank(amount)

    def set_background(self, background):
        self._background = background

    def as_dict(self):
        return {
            "ddd": self._ddd_name,
            "speaker": self._speaker,
            "modality": self._modality,
            "utterance": self._utterance,
            "understanding_confidence": self.understanding_confidence,
            "perception_confidence": self.perception_confidence,
            "move_type": self._type
        }

    def as_semantic_expression(self):
        return self._type


class MoveWithSemanticContent(Move, SemanticObjectWithContent):
    def __init__(self, type, content, *args, **kwargs):
        Move.__init__(self, type, *args, **kwargs)
        SemanticObjectWithContent.__init__(self, content)
        self._content = content

    @property
    def content(self):
        return self._content

    def get_content(self):
        return self._content

    def __hash__(self):
        return hash((
            self.__class__.__name__, self._type, self.perception_confidence, self.understanding_confidence,
            self.weighted_understanding_confidence, self._speaker, self._utterance, self._ddd_name
        ))

    def _get_semantic_expression(self, include_attributes):
        string = "Move("
        string += f"{self._type}({str(self._content)}"
        if self._background:
            string += f", {unicodify(self._background)}"
        string += ")"
        if include_attributes:
            string += self._build_string_from_attributes()
        string += ")"
        return string

    def as_dict(self):
        result = {
            "content": self.content.as_dict() if self.content else None,
        }
        return super().as_dict() | result

    def class_internal_move_content_equals(self, other):
        return self._content == other._content


class ICMMove(Move):
    RERAISE = "reraise"
    PER = "per"
    ACC = "acc"
    SEM = "sem"
    UND = "und"
    ACCOMMODATE = "accommodate"
    LOADPLAN = "loadplan"
    RESUME = "resume"
    REPORT_INFERENCE = "report_inference"
    CARDINAL_SEQUENCING = "cardinal_sequencing"

    INT = "int"
    POS = "pos"
    NEG = "neg"

    def __init__(
        self,
        type,
        understanding_confidence=None,
        speaker=None,
        polarity=None,
        ddd_name=None,
        perception_confidence=None
    ):
        if polarity is None:
            polarity = ICMMove.POS
        self._polarity = polarity
        Move.__init__(
            self,
            type,
            understanding_confidence=understanding_confidence,
            speaker=speaker,
            ddd_name=ddd_name,
            perception_confidence=perception_confidence
        )

    def class_internal_move_content_equals(self, other):
        return (self._polarity == other._polarity and self._type == other._type)

    def get_polarity(self):
        return self._polarity

    def is_icm(self):
        return True

    def is_issue_icm(self):
        return False

    def is_question_raising(self):
        return False

    def get_content(self):
        return None

    def is_negative_perception_icm(self):
        if self.type_ == ICMMove.PER:
            return self.get_polarity() == ICMMove.NEG
        else:
            return False

    def is_positive_acceptance_icm(self):
        if self.type_ == ICMMove.ACC:
            return self.get_polarity() == ICMMove.POS
        else:
            return False

    def is_negative_acceptance_issue_icm(self):
        return False

    def is_negative_acceptance_icm(self):
        if (self.type_ == ICMMove.ACC and self.get_polarity() == ICMMove.NEG):
            return True
        else:
            return False

    def is_negative_understanding_icm(self):
        return (self.type_ == ICMMove.UND and self.get_polarity() == ICMMove.NEG)

    def is_positive_understanding_icm_with_non_neg_content(self):
        return False

    def is_interrogative_understanding_icm_with_non_neg_content(self):
        return False

    def is_grounding_proposition(self):
        return False

    def __str__(self):
        return self.get_semantic_expression(include_attributes=True)

    def get_semantic_expression(self, include_attributes=True):
        string = f"ICMMove({self._icm_to_string()}"
        if include_attributes:
            if self._speaker:
                string += f", speaker={self._speaker}"
            if self.understanding_confidence is not None:
                string += f", understanding_confidence={self.understanding_confidence}"
            if self.perception_confidence is not None:
                string += f", perception_confidence={self.perception_confidence}"
        string += ")"
        return string

    def semantic_expression_without_realization_data(self):
        return self.get_semantic_expression(include_attributes=False)

    def _icm_to_string(self):
        if self._type in [ICMMove.PER, ICMMove.ACC, ICMMove.UND, ICMMove.SEM]:
            return f"icm:{self._type}*{self._polarity}"
        return f"icm:{self._type}"

    def as_semantic_expression(self):
        return self._icm_to_string()

    def as_dict(self):
        result = {"polarity": self.get_polarity()}

        return super().as_dict() | result


class IssueICMMove(ICMMove):
    def is_issue_icm(self):
        return True

    def is_negative_acceptance_issue_icm(self):
        if (self.type_ == ICMMove.ACC and self.get_polarity() == ICMMove.NEG):
            return True
        return False

    def _icm_to_string(self):
        return f"{ICMMove._icm_to_string(self)}:issue"


class ICMMoveWithContent(ICMMove):
    def __init__(self, type, content, content_speaker=None, *args, **kwargs):
        ICMMove.__init__(self, type, *args, **kwargs)
        self._content = content
        self._content_speaker = self._get_checked_content_speaker(content_speaker)

    @property
    def content(self):
        return self._content

    def get_content(self):
        return self._content

    def _get_checked_content_speaker(self, speaker):
        if (speaker in [Speaker.USR, Speaker.SYS, Speaker.MODEL, None]):
            return speaker
        raise Exception(f"'{speaker}' is not a valid value for content_speaker")

    def get_content_speaker(self):
        return self._content_speaker

    def _icm_to_string(self):
        if self._content_speaker is not None:
            return f"icm:{self._type}*{self._polarity}:{self._content_speaker}*{self._content}"

        if self._type == ICMMove.PER:
            return f'icm:{self._type}*{self._polarity}:"{self._content}"'
        if self._type in [ICMMove.ACC, ICMMove.UND, ICMMove.SEM]:
            return f"icm:{self._type}*{self._polarity}:{self._content}"
        return f"icm:{self._type}:{self._content}"

    def class_internal_move_content_equals(self, other):
        return (
            self._polarity == other._polarity and self._type == other._type and self._content == other._content
            and self._content_speaker == other._content_speaker
        )

    def is_question_raising(self):
        return (
            self.type_ == ICMMove.UND and self.content is not None
            and not (self.get_polarity() == ICMMove.POS and not self.content.is_positive())
        )

    def is_positive_understanding_icm_with_non_neg_content(self):
        return (self.type_ == ICMMove.UND and self.get_polarity() == ICMMove.POS and self.content.is_positive())

    def is_interrogative_understanding_icm_with_non_neg_content(self):
        return (self.type_ == ICMMove.UND and self.get_polarity() == ICMMove.INT and self.content.is_positive())

    def is_grounding_proposition(self):
        return self.type_ == ICMMove.UND and self.get_polarity() in [ICMMove.POS, ICMMove.INT]

    def as_dict(self):
        return super().as_dict() | {"content": self.content, "content_speaker": self.get_content_speaker()}


class CardinalSequencingICM(ICMMoveWithContent):
    def __init__(self, step):
        super().__init__(ICMMove.CARDINAL_SEQUENCING, step)

    def __eq__(self, other):
        try:
            return self.type == other.type and self._content == other._content
        except AttributeError:
            return False


class ICMMoveWithStringContent(ICMMoveWithContent):
    pass


class ICMMoveWithSemanticContent(ICMMoveWithContent, MoveWithSemanticContent):
    def __init__(
        self,
        type,
        content,
        understanding_confidence=None,
        speaker=None,
        ddd_name=None,
        content_speaker=None,
        polarity=None,
        perception_confidence=None
    ):
        MoveWithSemanticContent.__init__(
            self,
            type,
            content,
            understanding_confidence=understanding_confidence,
            speaker=speaker,
            ddd_name=ddd_name,
            perception_confidence=perception_confidence
        )
        ICMMoveWithContent.__init__(
            self,
            type,
            content,
            content_speaker=content_speaker,
            understanding_confidence=understanding_confidence,
            speaker=speaker,
            polarity=polarity,
            ddd_name=ddd_name,
            perception_confidence=perception_confidence
        )


class ReportMove(MoveWithSemanticContent):
    def __init__(self, content, *args, **kwargs):
        MoveWithSemanticContent.__init__(self, Move.REPORT, content, *args, **kwargs)

    def as_semantic_expression(self):
        return f"report({self.get_content().as_semantic_expression()})"

    def is_turn_yielding(self):
        return True

    def _get_semantic_expression(self, include_attributes):
        string = f"report({unicodify(self.get_content())}"
        if self._background:
            string += f", {unicodify(self._background)}"
        if include_attributes:
            string += self._build_string_from_attributes()
        string += ")"
        return string

    def class_internal_move_content_equals(self, other):
        return (self.get_content() == other.get_content())


class PrereportMove(Move, OntologySpecificSemanticObject):
    def __init__(self, ontology_name, service_action, arguments):
        self.service_action = service_action
        self.arguments = arguments
        Move.__init__(self, Move.PREREPORT)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    def get_service_action(self):
        return self.service_action

    def get_arguments(self):
        return self.arguments

    def __str__(self):
        return self.get_semantic_expression(include_attributes=True)

    def get_semantic_expression(self, include_attributes=True):
        string = f"prereport({self.service_action}, {unicodify(self.arguments)}"
        if include_attributes:
            string += self._build_string_from_attributes()
        string += ")"
        return string

    def semantic_expression_without_realization_data(self):
        return self.get_semantic_expression(include_attributes=False)

    def as_semantic_expression(self):
        return str(self)

    def class_internal_move_content_equals(self, other):
        return (self.service_action == other.service_action and self.arguments == other.arguments)

    def as_dict(self):
        return super().as_dict() | {
            "arguments": self.arguments,
            "service_action": self.service_action,
            "ontology_name": self.ontology_name
        }


class NoMove(Move):
    def __init__(self, *args, **kwargs):
        super(NoMove, self).__init__(Move.NO_MOVE, *args, **kwargs)


class GreetMove(Move):
    def __init__(self, *args, **kwargs):
        super(GreetMove, self).__init__(Move.GREET, *args, **kwargs)


class ThankYouMove(Move):
    def __init__(self, *args, **kwargs):
        super(ThankYouMove, self).__init__(Move.THANK_YOU, *args, **kwargs)


class ThankYouResponseMove(Move):
    def __init__(self, *args, **kwargs):
        super(ThankYouResponseMove, self).__init__(Move.THANK_YOU_RESPONSE, *args, **kwargs)


class InsultResponseMove(Move):
    def __init__(self, *args, **kwargs):
        super(InsultResponseMove, self).__init__(Move.INSULT_RESPONSE, *args, **kwargs)


class InsultMove(Move):
    def __init__(self, *args, **kwargs):
        super(InsultMove, self).__init__(Move.INSULT, *args, **kwargs)


class MuteMove(Move):
    def __init__(self, *args, **kwargs):
        super(MuteMove, self).__init__(Move.MUTE, *args, **kwargs)


class UnmuteMove(Move):
    def __init__(self, *args, **kwargs):
        super(UnmuteMove, self).__init__(Move.UNMUTE, *args, **kwargs)


class QuitMove(Move):
    def __init__(self, *args, **kwargs):
        super(QuitMove, self).__init__(Move.QUIT, *args, **kwargs)


class AskMove(MoveWithSemanticContent):
    def __init__(self, question, *args, **kwargs):
        MoveWithSemanticContent.__init__(self, Move.ASK, question, *args, **kwargs)

    def as_semantic_expression(self):
        return f"ask({self._content.as_semantic_expression()})"


class AnswerMove(MoveWithSemanticContent):
    def __init__(self, answer, *args, **kwargs):
        MoveWithSemanticContent.__init__(self, Move.ANSWER, answer, *args, **kwargs)

    def as_semantic_expression(self):
        return f"answer({self._content.as_semantic_expression()})"


class RequestMove(MoveWithSemanticContent):
    def __init__(self, action, *args, **kwargs):
        MoveWithSemanticContent.__init__(self, Move.REQUEST, action, *args, **kwargs)

    def as_semantic_expression(self):
        return f"request({self._content.as_semantic_expression()})"



import os
import shutil

DDD_MAKER_PATH = os.path.dirname(os.path.abspath(__file__))
TEMPLATES_PATH = os.path.join(DDD_MAKER_PATH, "templates")


def write_template_to_file(target_path, template):
    with open(target_path, 'w') as target:
        template.seek(0)
        shutil.copyfileobj(template, target)
